/*
 tools/evaltest.c
  evaluation tester

 To compile:
  C -o ::bin:evaltest :tools:evaltest.o 
   :obj:ftrig.a.o :obj:ftrig.c.o :obj:interpreter.c.o :obj:commandline.c.o

 Examples:
  ::bin:evaltest '-3<<+(-9)'
  ::bin:evaltest 'C*10.2' #the .2 is ignored
*/

#include <stdio.h>
#include <ctype.h>

#include "../macos/usercommands.h"

EXTERN_API(void)
SysBeep(short duration) ONEWORDINLINE(0xA9C8);

size_t strlen(const char*);

static evalvalue evalindirection(evalvalue v)
{
	return (v&0xFFFF)*12345;
}

static const char *getvalue(const char *w,
	const char *e,
	long int *i,
	unsigned char *u)
{
#pragma unused(u)
	if ((e=parse_wordbreak(w,e))==0)
		return 0;
	if (e-w==1 && toupper(*w)=='C')
	{
		*i = 299792458; /* light in a vaccuum, metres per second */
		return e;
	}

	return NULL;
}

static void paramiterator(unsigned a, unsigned char v)
{
(void)a;
	fputc(v,stdout);
}

static int beep_proc(int argc, char const*const*argv)
{
	long int	duration = 20;

	if (argc<=0)
	{ /* help message */
		if (++argc<=0)
			printf(
"%s [duration]\n%s",
				argv[0],argc<0 ?
" - beep for duration ticks (1/60 sec); default: 20\n":"");
		return 0;
	}

	if (--argc!=0)
		if (readanyinteger(argv[1],strlen(argv[1]),&duration)==NULL
				|| (unsigned)(duration-(-32768))>65535)
		{ /* help message */
			printf("%s: expected value within [-32768,32767]\n",
				argv[0]);
			return argc;
		}

	SysBeep(duration);

	return argc==1 ? 0:argc;
}

static int help_proc(int argc, char const*const*argv);

static const commandproc_keyword	keywords[] =
{
	{"beep", beep_proc},
	{"help", help_proc},
	{NULL, NULL}
};

static int help_proc(int argc, char const*const*argv)
{
	keyword_proc	proc;
	const char	*name = argv[0];

	if (argc<=0)
	{ /* help message */
		if (argc<0)
			printf("%s: display help for a command, or all (no parameter)\n",
				name);
		return 0;
	}

	if (--argc==0)
	{ /* no parameters; do all - will call this function */
		const commandproc_keyword	*keyword;

		keyword_proc	prev = NULL;

		printf("%s - showing all commands:\n",
			name);
		for(keyword=keywords; keyword->name!=NULL
				&& (proc=keyword->proc)!=NULL; prev=proc,++keyword)
			if (keyword->proc!=prev) /* ignore duplicates */
				proc(-1,&keyword->name);
		return 0;
	}

	do
	{
		int		length = strlen(*++argv);

		if ((proc=keywords_getproc(*argv,length,keywords))==NULL)
		{
			printf("%s: unable to find help for %s.\n",
				name,*argv);
			return argc;
		} else
			proc(-2,argv);
	} while(--argc!=0);

	return 0;
}

/* return negative for a proc for ~failed param, or +1 for too many parameters */
static int keyword_default(int argc, char const*const*argv)
{
	char		buf[24];
	const char	*name;

	if (argv==NULL || (name=*argv)==NULL)
		if (argc==EVAL_HUGEVAL)
		{
			printf("# evaltest: unrecognized command\n");
			return 1;
		} else
			name = "# evaltest";
	else if (argc==EVAL_HUGEVAL) /* blank */
	{
		printf("# evaltest: unrecognized '%s'\n",
			name);
		return 1;
	}

	if (--argc>=0)
	{
		if (argc!=0)
			printf("%s: %u parameter(s) were not used\n",
				name,argc);
		return 1;
	}
	if ((argc+=2)==0)
	{
		printf("%s: invalid parameter(s)\n",
			name);
		return 1;
	}
	if (argc<0)
	{
		printf("%s: error at parameter %u\n",
			name,-argc);
		return 1;
	}

	if (argv==NULL || *argv==NULL)
		{puts("# evaltest: unknown error\n"); return 1;}
	else
	{
		const char	*s = *argv;
		long int	value;
		long		length = strlen(s);

		if (*s=='=')
			++s; /* forced evaluation */
		if ((s=parse_parameter(s,length,buf,
				getvalue,NULL,
				evalindirection))==NULL)
			{printf("# evaltest: syntax error\n"); return 1;}
		if (buf[0]==' ')
			return 0; /* blank */

		s = parse_skipwsp(s,length+*argv);
		if (s-*argv!=length)
			if (buf[0]=='\0' && *s!='"' && *s!='\''
					&& (unsigned char)(*s-'0')>9)
			{
				printf("# evaltest - syntax error: '%s'\n",
					s);
				return 1;
			} else
				printf("# evaltest - what was not parsed: '%s'\n",
					s);

		if (readanyinteger(buf,strlen(buf),&value)==NULL)
		{
			printf("# evaltest: %s\n>",
				buf[0]=='\0' ? *argv:buf);
			parse_textloop(*argv,0,paramiterator);
			fputc('\n',stdout);
		} else
			printf(value<0 ?
				"# evaltest: '%.*s' = %s or %lu (0x%lX)\n":
				"# evaltest: '%.*s' = %s (0x%lX)\n",
				s-*argv,*argv,buf,(unsigned)value,(unsigned)value);
	}

	return 0;
}

int main(int argc, char **argv)
{
	const char	*s;
	int		z;

	if (argc<=0)
		return 0;
	if ((argc-=2)!=0 || (s=argv[1])==NULL)
	{
		fprintf(stderr,"### Usage: %s <expression>\n",
			argv[0]);
		return 1;
	}

	z = strlen(s);

	if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
		Debugger();

	if (commandline_execute(s,z,
			keywords,keyword_default,
			getvalue,NULL,
			evalindirection)!=NULL)
		;

	return 0;
}
