/*
 video.c

 * MacFCEU 0.6a4
 *
 * Copyright notice for this file:
 *  video.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <zlib.h>

#include "driver.h"
#include "FCEU.h"
#include "NSF.h"
#include "VSUni.h"
#include "video.h"
#include "general.h"
#include "memory.h"
#include "state.h"
#include "movie.h"
#include "palette.h"
#include "input.h"

#ifdef env68k
 #pragma segment shell
#endif

#define SHOWFPS
#define SNAP_IMAGE_WIDTH		256

uint8		*XBuf = 0;
static uint8	*xbsave = 0;

void FCEU_KillVirtualVideo(void)
{
	if (xbsave!=0)
	{
#ifdef mac
		if ((long int)xbsave>0) /* may be in address space (can't dispose) */
#endif
			FCEU_free(xbsave);
		xbsave = 0;
	}
}

int FCEU_InitVirtualVideo(void)
{
	if (XBuf==0) /* Some driver code may allocate XBuf externally. */
	{
		/* 256 bytes per scanline, * 256 scanline maximum, +16 for alignment */
		if ((XBuf=(uint8*)FCEU_malloc(256*256 + 16))==0)
			return 0;
	}

	xbsave = XBuf;

	XBuf = (uint8*)(~15&(15+(long int)XBuf));
	FCEU_memsetl(XBuf,0x80808080,256*256); //*240);

	return 1;
}

static int	howlong;
static char	errmsg[65];

static void ShowFPS(void);

#include "drawing.h"

static int dosnapsave = 0;
void FCEUI_SaveSnapshot(void)
{
	dosnapsave = 1;
}

static void ReallySnap(void)
{
	int		x;

	if ((x=SaveSnapshot())==0)
		FCEU_DispMessage("Error saving screen snapshot.");
	else
		FCEU_DispMessage("Screen snapshot %d saved.",
			x-1);
}

void FCEU_PutImage(void)
{
	if (FCEUGameInfo->type==GIT_NSF)
	{
		DrawNSF(XBuf);
	/* so we can see it; 2009.10.12 */
	#ifdef SHOWFPS
		ShowFPS();
	#endif
	/* Save snapshot after NSF screen is drawn.  Why would we want to do it before? */
		if (dosnapsave)
		{
			ReallySnap();
			dosnapsave = 0;
		}
	} else
	{
	#ifdef SHOWFPS
		ShowFPS();
	#endif
	/* Save snapshot before overlay stuff is written. */
		if (dosnapsave)
		{
			ReallySnap();
			dosnapsave = 0;
		}
		if (FCEUGameInfo->type==GIT_VSUNI)
			FCEU_VSUniDraw(XBuf,0x01);
		FCEU_DrawSaveStates(XBuf);
		FCEU_DrawMovies(XBuf);
		FCEU_DrawNTSCControlBars(XBuf);
	}

	DrawMessage();
	FCEU_DrawInput(XBuf);
}

/* made const char*; 2010.01.05 */
void FCEU_DispMessage(const char *format, ...)
{
	va_list		ap;

	va_start(ap,format);
	vsprintf(errmsg,format,ap);
	va_end(ap);

	howlong = 180;
}

void FCEU_ResetMessages(void)
{
	howlong = 0;
}


static int WritePNGChunk(FILE *fp,
	uint32 size,
	char *type,
	uint8 *data)
{
	uint8		tempo[4];
	uint32		crc;

	tempo[0] = size>>24;
	tempo[1] = size>>16;
	tempo[2] = size>>8;
	tempo[3] = size;

	if (fwrite(tempo,4,1,fp)!=1)
		return 0;
	if (fwrite(type,4,1,fp)!=1)
		return 0;

	if (size!=0)
		if (fwrite(data,1,size,fp)!=size)
			return 0;

	crc = FCEUI_CRC32(0,(uint8 *)type,4);
	if (size!=0)
		crc = FCEUI_CRC32(crc,data,size);

	tempo[0] = crc>>24;
	tempo[1] = crc>>16;
	tempo[2] = crc>>8;
	tempo[3] = crc;

	return fwrite(tempo,4,1,fp)!=1 ? 0:1;
}

static int WritePNGHeader(FILE *pp,
	int width,
	int height,
	int bitdepth)
{
	static uint8	header[8] = {137, 80, 78, 71, 13, 10, 26, 10};
	uint8		chunko[13];

	if (fwrite(header,8,1,pp)!=1)
		return 0;

	chunko[0] = chunko[1] = 0;
	chunko[2] = (width>>8)&0xFF;
	chunko[3] = width&0xFF;

	chunko[4] = chunko[5] = 0;
	chunko[6] = (height>>8)&0xFF;
	chunko[7] = height&0xFF;

	chunko[8] = bitdepth;
	chunko[9] = bitdepth<=8 ? 3:0; // color type (indexed or direct?)
	chunko[10] = 0;	// deflate method of compression
	chunko[11] = 0;	// basic adapative filter set (though none are used)
	chunko[12] = 0;	// not interlaced

	return WritePNGChunk(pp,13,"IHDR",chunko);
}

static int WritePNGPalette(FILE *pp)
{
	uint8		pdata[256*3], *pptr = pdata;
	int		x;

	x = 0;
	do
		FCEUD_GetPalette(x,pptr,1+pptr,2+pptr);
	while(pptr+=3,++x<256);

	return WritePNGChunk(pp,256*3,"PLTE",pdata);
}

/* important: starts with filter code byte (expects (1+rowsize)*height) */
static int GetPNGImage(void *pixels,
	int rowsize,
	int height)
{
	uint8		*s = XBuf+FSettings.FirstSLine*rowsize;
	uint8		*d = pixels;
	int		x, y = height;

	if ((x=rowsize)<=0 || y<=0)
		return (x|y)<0 ? -1:0;
	do
	{
		*d++ = 0; // no filter (initially uncompressed)
		if (rowsize&1)
			*d++ = *s++;
		x = (unsigned)rowsize>>2; /* unaligned, four at a time */
		if (rowsize&2)
			goto two;
		do
		{
			*d++ = *s++, *d++ = *s++;
		two:	*d++ = *s++, *d++ = *s++;
		} while(--x!=0);
	} while(--y!=0);

	return 1;
}

int SaveSnapshot(void)
{
	static unsigned int	lastu = 0;

	char		*fn = 0;
	int		linecount = 1+FSettings.LastSLine-FSettings.FirstSLine;
	int		ix;
	FILE		*pp = 0;
	uint8		*compmem = 0;
	uLongf		compmemsize = linecount*263+12;

	if ((compmem=(uint8*)FCEU_malloc(compmemsize))==0)
		return 0;
	/* find an available file name */
	for(ix=lastu; ix<32767; ++ix) /* was 99999 */
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_SNAP,ix,"png"))==0
				|| (pp=FCEUD_UTF8fopen(fn,"rb"))==0)
			break;
		fclose(pp);
	}

	lastu = ix;

	pp = FCEUD_UTF8fopen(fn,"wb");
	free(fn);
	if (pp==0)
		return 0;

	if (WritePNGHeader(pp,SNAP_IMAGE_WIDTH,linecount,8)<=0)
		goto PNGerr;

	if (WritePNGPalette(pp)<=0)
		goto PNGerr;

	{
		uint8		*pixels;

		if ((pixels=(uint8 *)FCEU_malloc(linecount*SNAP_IMAGE_WIDTH+linecount))==0)
			goto PNGerr;

		GetPNGImage(pixels,SNAP_IMAGE_WIDTH,linecount);

		if (compress(compmem,&compmemsize,pixels,
				linecount*SNAP_IMAGE_WIDTH+linecount)!=Z_OK)
		{
			if (pixels!=0)
				free(pixels);
			goto PNGerr;
		}

		if (pixels!=0)
			FCEU_free(pixels);
		if (WritePNGChunk(pp,compmemsize,"IDAT",compmem)==0)
			goto PNGerr;
	}

	if (WritePNGChunk(pp,0,"IEND",0)==0)
	{
	PNGerr:
		ix = -1;
	}

	if (compmem!=0)
		FCEU_free(compmem);
	if (pp!=0)
		fclose(pp);

	return ix+1;
}

uint64		videofps_boop[60];

#ifdef SHOWFPS
uint64 FCEUD_GetTime(void);
uint64 FCEUD_GetTimeFreq(void);

static int	boopcount = 0;

 #ifdef mac
  #include "utility/ftrig.h"

uint8		show_fps;
 #endif

static void ShowFPS(void)
{
	char		fpsmsg[16];
	uint64		da = FCEUD_GetTime();
 #ifdef mac
	char		*fpsmsp = fpsmsg+15;
  #define booplimit 10
	if (!show_fps)
		goto end;
 #else
	int		booplimit = PAL?50:60;
 #endif

	(void)uint64sub(da,videofps_boop[boopcount]);
	videofps_boop[boopcount] = FCEUD_GetTime();

 #ifdef mac
	*fpsmsp = 0;
	if (int64z(videofps_boop[0]) || int64z(videofps_boop[9])
			|| int64hi(da)!=0 || int64lo(da)==0)
	*--fpsmsp = '-', *--fpsmsp = '-';
else
	{ /* always in microseconds */
		uint32	tenf = (4294967295.0+10)/10;
		uint32	numer = (20000000UL*booplimit)/int64lo(da)+1>>1;
		uint32	quot = ufract32mul(numer,tenf);
		/*sprintf(fpsmsg,"%u.%c", quot,perdec-quot*10+'0');*/
		*--fpsmsp = '0'+numer-quot*10;
		*--fpsmsp = '.';
		numer = quot;
		do
			quot = ufract32mul(numer,tenf), *--fpsmsp = '0'+numer-quot*10;
		while((numer=quot)!=0 && fpsmsp>fpsmsg);
	}
  #define fpsmsg fpsmsp
 #else
	/* double divide */
	//sprintf(fpsmsg, "%8.1f",(double)booplimit / ((double)da / FCEUD_GetTimeFreq()));
	sprintf(fpsmsg,"%8.1f", FCEUD_GetTimeFreq()*(double)booplimit/(double)da);
 #endif

	if ((unsigned)FSettings.FirstSLine<=240)
		DrawTextTrans(XBuf+(256-8-8*8)+(FSettings.FirstSLine+3)*256,256,(uint8*)fpsmsg,4);
	// It's not averaging FPS over exactly 1 second, but it's close enough.
end:
	boopcount = (boopcount+1)%booplimit;
}
#endif /*defined(SHOWFPS)*/
